#!/usr/bin/env python
# vim: set fileencoding=UTF-8 :

#Libraries
from l_config import config
# import l_analog
# import l_drive
import smbus
from datetime import datetime
import time

import numpy as np

import multiprocessing as mp
import ctypes

class gyro_asynch:
    def __init__(self):
        self.bus = smbus.SMBus(1)
        self.port = 0x68
        self.wakeup()
        #self.reset_device()
        
        self.t_label=mp.Value(ctypes.c_double, 0.0)
        self.yaw=mp.Value(ctypes.c_double, 0.0)
        self.yaw_dot=mp.Value(ctypes.c_double, 0.0)
        manager = mp.Manager()
        self.hist=manager.list()
        self.start_flag=mp.Value(ctypes.c_int, 1)
    def start(self,f=False):    
        self.p = mp.Process(target=self.gyrosensor, args=[self.t_label,self.yaw,self.yaw_dot,self.hist,self.start_flag,f])
        self.p.start()
        #self.p.join()

    def end(self):
        #self.p.terminate()
        self.start_flag.value=0

    def kill(self):
        self.p.terminate()
    def wakeup(self):    
        self.bus.write_byte_data(self.port, 0x6b, 0)
        
    def reset_device(self):
        from pycomms import PyComms
        pc=PyComms(self.port)
        pc.writeBit(0x6b, 7,True)
        
    def read_byte(self,adr):
        return self.bus.read_byte_data(self.port, adr)
    
    def read_word(self,adr):
        high = self.bus.read_byte_data(self.port, adr)
        low = self.bus.read_byte_data(self.port, adr+1)
        val = (high << 8) + low
        return val
    
    def read_word_2c(self,adr):
        val = self.read_word(adr)
        if (val >= 0x8000):
            return -((65535 - val) + 1)
        else:
            return val
    
    def gyrosensor(self,t,yaw,yaw_dot,hist_shared,start_flag,f=False):  
        
        #Parametter
        k=config.gyro.multiplicator*config.gyro.fine_gain
        a=config.gyro.zero_balance
        sgz=0
        s_old=0
        i=1
        t_arr=[]
        yaw_arr=[]
        yaw_dot_arr=[]
        start = datetime.now()    
            
        while True:
            now = datetime.now()
            gyro_zout = self.read_word_2c(0x47)
            delta = now - start
            s = delta.seconds + delta.microseconds / 1E6

            gz= -(gyro_zout+a)/(k)
            
            if i>1:
                #print (s-s_old)
                sgz+=gz*(s-s_old)
                t.value=s
                yaw.value=sgz
                yaw_dot.value=gz
                t_arr.append(s)
                yaw_arr.append(sgz)
                yaw_dot_arr.append(gz)
            s_old=s
            i+=1
            if f:
                #print "Sleeping"
                time.sleep(1./float(f))
                
            if start_flag.value==0:
                print "Gyro stopped"
                hist_shared[:]=[t_arr,yaw_arr,yaw_dot_arr]
                break
                sys.exit()

        