#!/usr/bin/env python
# vim: set fileencoding=UTF-8 :

#Libraries
from l_config import config
import l_gyro
import l_odo
import l_config
import smbus
import math
from datetime import datetime
import time
import numpy as np

import multiprocessing as mp
import ctypes

class localize_asynch:
    def __init__(self):
        config=l_config.config
        self.d=config.hardware.wheel_distance    
        self.t_label=mp.Value(ctypes.c_double, 0.0)
        manager = mp.Manager()
        self.y=manager.list()
        self.signs=manager.list()
        self.t=manager.list()
    def start(self,f=False):    
        self.p = mp.Process(target=self.localize, args=[self.t,self.y,self.signs])
        self.p.start()
        #self.p.join()

    def kill(self):
        self.p.terminate()

    def set_odo_signs(self,sgn):
        
        self.signs[:]=sgn

    def get_arc_lengths(self,d_alpha,r):
        d=self.d
        if d_alpha>0:
            enc_l=(r+d/2)*d_alpha
            enc_r=(r-d/2)*d_alpha
        elif d_alpha<0:
            enc_l=(r-d/2)*d_alpha
            enc_r=(r+d/2)*d_alpha
        elif d_alpha==0:
            print "Something went wrong"
        return [enc_l,enc_r]
    
    #Get the displacement by only odo
    def get_arc(self,d_a_l,d_a_r):

        if d_a_l!=d_a_r:
            r=(self.d*(d_a_l+d_a_r))/(2.*(d_a_l-d_a_r))
            
            d_alpha_hat=(d_a_l-d_a_r)/self.d
            dx_hat=r*math.sin(d_alpha_hat)
            dy_hat=r*(1-math.cos(d_alpha_hat))
        else:
            r=float("inf")
            dx_hat=d_a_l
            dy_hat=0
            d_alpha_hat=0
        
        if False:
            #print "################################   Turn ",i," t=",t_label,"######################################"
            #print "Actual enc_l= ",enc_l[-1]," Actual enc_r= ",enc_r[-1]
            print "Diff enc_l= ",d_a_l," Diff enc_r= ",d_a_r
            print "Actual radius= ",r
            print "Diff arc angle= ",d_alpha_hat,"(",math.degrees(d_alpha_hat),")", "Diff x local= ",dx_hat," Diff y local= ",dy_hat
            print "-------------Kinematics equations-------------"
        
        #Local coordinate system
        return [r,d_alpha_hat]
    
    
    def get_local_disp(self,r,d_alpha,d_a_l):
        #print "Alpha:",d_alpha
        if d_alpha!=0:
            dx_hat=r*math.sin(d_alpha)
            dy_hat=r*(1-math.cos(d_alpha))
        else:
            dx_hat=d_a_l
            dy_hat=0
    
        return np.matrix([dx_hat,dy_hat,d_alpha]).transpose()

    def get_radius_range(self,alpha,enc_l,enc_r,eps_odo):

        d=self.d
        if alpha>0:        
            enc_l_max=enc_l+eps_odo
            enc_r_min=enc_r-eps_odo
             
            r_max=enc_l_max/alpha-d/2
            r_min=enc_r_min/alpha+d/2
     
        elif alpha<0:
            enc_l_min=enc_l-eps_odo
            enc_r_max=enc_r+eps_odo
             
            r_max=enc_r_max/alpha-d/2
            r_min=enc_l_min/alpha+d/2
        elif alpha==0:
            r_max=False
            r_min=False
        return [r_min,r_max]
    
    def get_line_range(self,enc_l,enc_r,eps_odo):
        enc_l_max=enc_l+eps_odo
        enc_l_min=enc_l-eps_odo
        enc_r_max=enc_r+eps_odo
        enc_r_min=enc_r-eps_odo
        
        enc_min=max(enc_l_min,enc_r_min)
        enc_max=min(enc_l_max,enc_r_max)
        
        return [enc_min,enc_max]
    
    def get_local_disp(self,r,d_alpha,d_a_l):
        #print "Alpha:",d_alpha
        if d_alpha!=0:
            dx_hat=r*math.sin(d_alpha)
            dy_hat=r*(1-math.cos(d_alpha))
        else:
            dx_hat=d_a_l
            dy_hat=0
    
        return np.matrix([dx_hat,dy_hat,d_alpha]).transpose() 
    
    def update_global_disp(self,u_glob,du_local):
        #Rotation of the local system regarding the global system 
        #print du_local
        #print u_glob
        d_alpha_gl=u_glob[2,-1]
        #Transformation-matrix (Global to local)
        T= np.matrix([[math.cos(d_alpha_gl), -math.sin(d_alpha_gl),0],
                      [math.sin(d_alpha_gl), math.cos(d_alpha_gl),0],
                      [0,0,1]])
        #Displacement in the global system
        du_glob=T*du_local
        #Absolute position
        u_glob=np.concatenate((u_glob,u_glob[:,-1]+du_glob),axis=1)
        #print u_glob
        return u_glob

    
    def localize(self,t,y,signs):      
        #Start processes
        #Gyrometer
        gyro=l_gyro.gyro_asynch()
        gyro.start(1000)
        #Odometry
        odo=l_odo.odo_asynch() #Define object
        odo.start(1000) #Start measuring with the frequency
        
        t_global=[0]
        u_ref=np.matrix([0,0,0]).transpose()
        u_global_predicted=np.matrix([0,0,0]).transpose()
        u_global_odo=np.matrix([0,0,0]).transpose()
        enc_l_read_arr=[odo.rad_to_arc(odo.enc_l.value)]
        enc_r_read_arr=[odo.rad_to_arc(odo.enc_r.value)]
        enc_l_arr=[odo.rad_to_arc(odo.enc_l.value)]
        enc_r_arr=[odo.rad_to_arc(odo.enc_r.value)]
        yaw_vector=[0]
        eps_lin=0.005
        eps_odo=odo.rad_to_arc(1)
        #eps_odo=0
        x_dot_arr=np.matrix([0,0,0]).transpose()
        #odo.set_sign(signs)
        #print odo
        
        #Control circuit
        while True:
            ####################################
            #Position prediction
            ####################################
            #Time label of the measurement
            t[:]=[odo.t_label.value] 
            t_global.append(t[0])
            #print t
            #Odometer
            #Odo sign
            odo.set_sign(self.signs)
            
            enc_l_read_arr.append(odo.rad_to_arc(odo.enc_l.value))
            enc_r_read_arr.append(odo.rad_to_arc(odo.enc_r.value))
            #print "Read left:",enc_l_read_arr[-1],"Read right:",enc_r_read_arr[-1]
            #Read gyro
            yaw_vector.append(math.radians(gyro.yaw.value))
            d_yaw=yaw_vector[-1]-yaw_vector[-2]
            
            #print "####"
            #print "enc ", enc_l_read_arr[-1],enc_r_read_arr[-1]
            #print "enc pre ", enc_l_arr[-1], enc_r_arr[-1]
            #print "#####"
            #Get the displacement in the local coordinates: Only by Odo for check reasons
            if False:
                #print enc_l_read_arr[-1]-enc_l_read_arr[-2],enc_r_read_arr[-1]-enc_r_read_arr[-2]
                [r_odo,d_alpha_odo]=self.get_arc(enc_l_read_arr[-1]-enc_l_read_arr[-2],enc_r_read_arr[-1]-enc_r_read_arr[-2])
                #print "Arc odo",[r_odo,d_alpha_odo]
                
    
                if d_alpha_odo!=0:
                    du_local_odo=self.get_local_disp(r_odo,d_alpha_odo,False)
                elif d_alpha_odo==0:
                    du_local_odo=self.get_local_disp(0,d_alpha_odo,enc_l_read_arr[-1]-enc_l_read_arr[-2])
                    
                u_global_odo=self.update_global_disp(u_global_odo,du_local_odo)
                    
                #print "odo glob ",u_global_odo[:,-1]
                
            if d_yaw!=0:            
                #Radius range
                 
                [r_min,r_max]=self.get_radius_range(d_yaw,enc_l_read_arr[-1]-enc_l_arr[-1],enc_r_read_arr[-1]-enc_r_arr[-1],eps_odo)
                #[r_min,r_max]=self.get_radius_range(d_yaw,enc_l_read_arr[-1]-enc_l_read_arr[-2],enc_r_read_arr[-1]-enc_r_read_arr[-2],eps_odo)
                #Predict radius        
                r_predicted=(r_max+r_min)/2
                #r_predicted=r_min
                #Get predicted arc lengths
                [enc_l_predicted,enc_r_predicted]=self.get_arc_lengths(d_yaw,r_predicted)
                #print "r_min=%.3f\tr_max=%.3f\tR Predicted:%.3f"%(r_min,r_max,r_predicted)
                #Get local displacements by this arcs
                #print "lpre ",enc_l_predicted,"rpre ",enc_r_predicted
                du_local_predicted=self.get_local_disp(r_predicted,d_yaw,False)
                       
            elif d_yaw==0:
                #Prediction
                #Radius range
                [l_min,l_max]=self.get_line_range(enc_l_read_arr[-1]-enc_l_arr[-1],enc_r_read_arr[-1]-enc_r_arr[-1],eps_odo)
                #Predict line length         
                enc_l_predicted=(l_max+l_min)/2
                enc_r_predicted=enc_l_predicted
                #Get local displacements by this arcs
                du_local_predicted=self.get_local_disp(0,0,enc_l_predicted)
                
            #Predicted encoder arrays
                                   
            enc_l_arr.append(enc_l_arr[-1]+enc_l_predicted)
            enc_r_arr.append(enc_r_arr[-1]+enc_r_predicted)
       
            
            #################################
            #Reconstruct in the global system
            #################################
            u_global_predicted=self.update_global_disp(u_global_predicted,du_local_predicted)
            #print "odo predicted ",u_global_predicted[:,-1]
            #Update the state
            x=u_global_predicted[:,-1]
            #Derivative

            
            
            if (t_global[-1]-t_global[-2]!=0) and not np.array_equal(u_global_predicted[:,-1],u_global_predicted[:,-2]):
#                 print "Set"
#                 print u_global_predicted[:,-1]
#                 print u_global_predicted[:,-2]
                x_dot=(u_global_predicted[:,-1]-u_global_predicted[:,-2])/(t_global[-1]-t_global[-2])
                
                
            else:
#                 print "Here"
                x_dot=x_dot_arr[:,-1]
                
            x_dot_arr=np.concatenate((x_dot_arr,x_dot),axis=1)
            #Floating mean filter
            max_depth=5
            #print "shape",x_dot_arr.shape[1]
            if x_dot_arr.shape[1]<max_depth:
                depth=x_dot_arr.shape[1]
            else:
                depth=max_depth
            
            
            arr_cut=x_dot_arr[:,-depth:-1]
            #print "arr_cut", arr_cut
            filtered=x_dot_arr.sum(axis=1)/x_dot_arr.shape[1]
            #print "#######Filtered\n",filtered,"\n###########"
            
            #print "x_dot ",x_dot
            y[:]= np.concatenate((x,x_dot),axis=0).transpose().tolist()[0]
            time.sleep(0.01)
