#!/usr/bin/python
#############################################################
#############################################################
# Auto Drive
# Author: Can Kosar
#############################################################

#Standard libraries
import sys
import os
import time
import math
import numpy as np
from scipy import integrate
import matplotlib.pyplot as plt
import sys


#Custom libraries
sys.path.append("../lib/")
import l_drive
import l_odo
import l_config
import l_monitoring
import l_misc
#import l_accel
import l_gyro
#Initializing
config=l_config.config
d=config.hardware.wheel_distance

#Get the displacement caused by arc motion
def get_arc(d_a_l,d_a_r):
    if d_a_l!=d_a_r:
        r=(d*(d_a_l+d_a_r))/(2.*(d_a_l-d_a_r))
        
        d_alpha_hat=(d_a_l-d_a_r)/d
        dx_hat=r*math.sin(d_alpha_hat)
        dy_hat=r*(1-math.cos(d_alpha_hat))
    else:
        r=float("inf")
        dx_hat=d_a_l
        dy_hat=0
        d_alpha_hat=0
    
    if True:
        #print "################################   Turn ",i," t=",t_label,"######################################"
        #print "Actual enc_l= ",enc_l[-1]," Actual enc_r= ",enc_r[-1]
        print "Diff enc_l= ",d_a_l," Diff enc_r= ",d_a_r
        print "Actual radius= ",r
        print "Diff arc angle= ",d_alpha_hat,"(",math.degrees(d_alpha_hat),")", "Diff x local= ",dx_hat," Diff y local= ",dy_hat
        print "-------------Kinematics equations-------------"
    
    #Local coordinate system
    return [r,d_alpha_hat]

def get_arc_lengths(d_alpha,r):
    if d_alpha>0:
        enc_l=(r+d/2)*d_alpha
        enc_r=(r-d/2)*d_alpha
    elif d_alpha<0:
        enc_l=(r-d/2)*d_alpha
        enc_r=(r+d/2)*d_alpha
    elif d_alpha==0:
        print "Something went wrong"
    return [enc_l,enc_r]

def get_local_disp(r,d_alpha,d_a_l):
    #print "Alpha:",d_alpha
    if d_alpha!=0:
        dx_hat=r*math.sin(d_alpha)
        dy_hat=r*(1-math.cos(d_alpha))
    else:
        dx_hat=d_a_l
        dy_hat=0

    return np.matrix([dx_hat,dy_hat,d_alpha]).transpose()
    

def get_radius_range(alpha,enc_l,enc_r,eps_odo):
    
    print "EncL",enc_l
    print "EncR",enc_r
    if alpha>0:        
        enc_l_max=enc_l+eps_odo
        enc_r_min=enc_r-eps_odo
        
        r_max=enc_l_max/alpha-d/2
        r_min=enc_r_min/alpha+d/2

    elif alpha<0:
        enc_l_min=enc_l-eps_odo
        enc_r_max=enc_r+eps_odo
        
        r_max=enc_r_max/alpha-d/2
        r_min=enc_l_min/alpha+d/2
    elif alpha==0:
        r_max=False
        r_min=False
        
    return [r_min,r_max]

def get_line_range(enc_l,enc_r,eps_odo):
    enc_l_max=enc_l+eps_odo
    enc_l_min=enc_l-eps_odo
    enc_r_max=enc_r+eps_odo
    enc_r_min=enc_r-eps_odo
    
    enc_min=max(enc_l_min,enc_r_min)
    enc_max=min(enc_l_max,enc_r_max)
    
    return [enc_min,enc_max]
    
def update_global_disp(u_glob,du_local):
        #Rotation of the local system regarding the global system 
        #print du_local
        #print u_glob
        d_alpha_gl=u_glob[2,-1]
        #Transformation-matrix (Global to local)
        T= np.matrix([[math.cos(d_alpha_gl), -math.sin(d_alpha_gl),0],
                      [math.sin(d_alpha_gl), math.cos(d_alpha_gl),0],
                      [0,0,1]])
        #Displacement in the global system
        du_glob=T*du_local
        #Absolute position
        u_glob=np.concatenate((u_glob,u_glob[:,-1]+du_glob),axis=1)
        #print u_glob
        return u_glob



def co_simulation_straight(y_0):
    global sim_a_t_s,sim_u_s,sim_u_dot_s
    delay=0.001
    
    def ode_equ(y, t):
        m=config.hardware.mass
        r=config.hardware.r_robot
        d=9
        return [y[1], -d/m*y[1]]
    
    
    #Delay
    a_t_delay=[0,delay]
    sim_u_dot_delay=[y_0[1],y_0[1]]    
    sim_u_delay=[y_0[0],y_0[0]+delay*y_0[1]]
    
    #Damping
    a_t = np.arange(0+delay, 1.0+delay, 0.2)
    y_after_delay=[sim_u_delay[-1],sim_u_dot_delay[-1]]    
    asol = integrate.odeint(ode_equ, y_after_delay, a_t)
    
    #print a_t_delay
    #print sim_u_delay
    #print sim_u_dot_delay
    #sys.exit()
    sim_a_t_s=np.concatenate((a_t_delay,a_t)).tolist()
    sim_u_s=np.concatenate((sim_u_delay,asol[:,0])).tolist()
    sim_u_dot_s=np.concatenate((sim_u_dot_delay,asol[:,1])).tolist()

    return sim_u_s[-1]



def control_straight(dist,u_global):
#     global gyro
#     global odo

    #Drive       
    drive=l_drive.drive()
    duty=70
    balance_calibration=17
    sgn=drive.move_fw(duty,balance_calibration)
    odo.set_sign(sgn)

    #Controlling
    a_l_last=0
    a_r_last=0
    t_global=[0]
    u_ref=np.matrix([0,0,0]).transpose()
    u_global_predicted=u_global
    enc_l_read_arr=[odo.rad_to_arc(odo.enc_l.value)]
    enc_r_read_arr=[odo.rad_to_arc(odo.enc_r.value)]
    enc_l_arr=[odo.rad_to_arc(odo.enc_l.value)]
    enc_r_arr=[odo.rad_to_arc(odo.enc_r.value)]
    yaw_vector=[0]
    eps_lin=0.005
    eps_odo=odo.rad_to_arc(1)
    fo_start=False
    fo_time=0.5
    print "######################Main loop###################################"
    print "#                                                                #"
    print "#                                                                #"
    print "#                                                                #"
    print "#                                                                #"
    print "##################################################################"
    i=1
    
    
    #Control circuit
    while True:
        ####################################
        #Position prediction
        ####################################
        #Time label of the measurement
        t_label=odo.t_label.value 
        #Odometer
        enc_l_read_arr.append(odo.rad_to_arc(odo.enc_l.value))
        enc_r_read_arr.append(odo.rad_to_arc(odo.enc_r.value))
        print enc_r_read_arr
        print enc_r_read_arr
        #Read gyro
        yaw_vector.append(math.radians(gyro.yaw.value))
        d_yaw=yaw_vector[-1]-yaw_vector[-2]
        #yaw_dot=gyro.yaw_dot.value
        t_global.append(t_label)
        #Get the displacement in the local coordinates: Only by Odo for check reasons
        [r_odo,d_alpha_odo]=get_arc(enc_l_read_arr[-1]-enc_l_read_arr[-2],enc_r_read_arr[-1]-enc_r_read_arr[-2])
        #print "Arc odo",[r_odo,d_alpha_odo]
        if d_alpha_odo!=0:
            du_local=get_local_disp(r_odo,d_alpha_odo,False)
        elif d_alpha_odo==0:
            du_local=get_local_disp(0,d_alpha_odo,enc_l_read_arr[-1]-enc_l_read_arr[-2])
            
        if True:
            #Prediction
            print "#############\nPredicted"
            if d_yaw!=0:            
                #Radius range
                [r_min,r_max]=get_radius_range(d_yaw,enc_l_read_arr[-1]-enc_l_arr[-1],enc_r_read_arr[-1]-enc_r_arr[-1],eps_odo)
                #Predict radius        
                r_predicted=(r_max+r_min)/2
                #Get predicted arc lengths
                [enc_l_predicted,enc_r_predicted]=get_arc_lengths(d_yaw,r_predicted)
                print "r_min=%.3f\tr_max=%.3f\tR Predicted:%.3f"%(r_min,r_max,r_predicted)
                #Get local displacements by this arcs
                du_local_predicted=get_local_disp(r_predicted,d_yaw,False)
                
            elif d_yaw==0:
                #Prediction
                #Radius range
                [l_min,l_max]=get_line_range(enc_l_read_arr[-1]-enc_l_arr[-1],enc_r_read_arr[-1]-enc_r_arr[-1],eps_odo)
                #Predict line length         
                enc_l_predicted=(l_max+l_min)/2
                enc_r_predicted=enc_l_predicted
                #Get local displacements by this arcs
                du_local_predicted=get_local_disp(0,0,enc_l_predicted)
                            #Predicted encoder arrays
                            
            enc_l_arr.append(enc_l_arr[-1]+enc_l_predicted)
            enc_r_arr.append(enc_r_arr[-1]+enc_r_predicted)
            
            #print "Predicted local at dyaw=%.3f"%d_yaw
            #print du_local_predicted
        
        #################################
        #Reconstruct in the global system
        #################################
        
        u_global=update_global_disp(u_global,du_local)
        u_global_predicted=update_global_disp(u_global_predicted,du_local_predicted)
        ##############
        #Controlling
        ##############
        #Error vector (In this case it's y-displacement: e(t)=u_G,y(t)
        if True:
            
            u_ref=update_global_disp(u_ref,du_local)
            #print u_ref
            e=u_ref[:,-1]
            t=t_global[-1]
            if t-t_global[-2]>0:
                e_dot=(e-u_ref[1,-2])/(t-t_global[-2])
            else:
                e_dot=np.matrix([0,0,0]).transpose()
                print "Warning: Time increment unsuccessfull"
            
            #State space
            ss=np.concatenate((e,e_dot),axis=0).transpose()
            #Parameter matrix    
            A=np.matrix([0,-1000,-150,0,-1,-0.5]).transpose()
            #Set matrix
            #print l_misc.plots().print_matrix(ss)
            #print l_misc.plots().print_matrix(A)
            sm=ss*A       
            
            #PD-Controller
            #print l_misc.plots().print_matrix(e)        
            new_balance=sm[0]        
            #new_balance=0
            
            if not fo_start:
                sgn=drive.move_fw(duty,balance_calibration+new_balance)
            
        #############################################
        #Monitoring#
        #############################################
        #Monitoring
        print "################################   Turn ",i," t=",t_label,"######################################"
        if False:
            print "-------Controlling---------"
            print "yaw"
            print yaw_vector[-1]
            print "e"
            print l_misc.plots().print_matrix(e)
            print "e dot"
            print l_misc.plots().print_matrix(e_dot)
            print "New balance=%.3f"%new_balance
            print "---------------------------"
            print "du_local"
            print l_misc.plots().print_matrix(du_local)
            print "T"
            print l_misc.plots().print_matrix(T)
            print "du_global"
            print l_misc.plots().print_matrix(du_global)
            print "u_global"
            print l_misc.plots().print_matrix(u_global)
            print "\n\n\n"
        #############################################
        #Exit criteria    
        #Co-Simulation
        u1=u_ref[0,-1]
        du1=(u_ref[0,-1]-u_ref[0,-2])/(t_global[-1]-t_global[-2])        
        rm=co_simulation_straight([u1,du1])
        
        print "x=%.3f\tdx=%.3f\trm=%.3f\t"%(u1,du1,rm)
        #Break if x exceeds the value
        if math.fabs(rm)>=math.fabs(dist)-eps_lin and not fo_start:
            print "Breaking"
            drive.break_move()            
            print "##############################"
            fo_start=t_label
                
        if fo_start:
            if t_label-fo_start>=fo_time:
                #gyro.end()
                time.sleep(0.5)
                break
     
        
        
        #Clocking the controlling
        i+=1
        time.sleep(0.1)
    
    #gyro.kill()
    #odo.kill()   
    
    
    #############################################
    #Monitoring#
    #############################################
    if True:
        print "Yaw:%.6f"%yaw_vector[-1]
        print "Global position Odo-only"
        print u_global[:,-1]
        print "Global position Predicted"
        print u_global_predicted[:,-1]
    #One piece arc
    if False:
        du_global_op=get_local_displacement(enc_l[-1],enc_r[-1])
        print "Global position one piece arc"
        print du_global_op

    return [t_global,u_global_predicted,yaw_vector]

def co_simulation_turn(y_0):
    global sim_a_t,sim_u,sim_u_dot
    delay=0.001
    
    def ode_equ(y, t):
        m=config.hardware.mass
        r=config.hardware.r_robot
        Iz=(m*(r**2))/2
        d=0.18
        return [y[1], -d/Iz*y[1]]
    
    
    #Delay
    a_t_delay=[0,delay]
    sim_u_dot_delay=[y_0[1],y_0[1]]    
    sim_u_delay=[y_0[0],y_0[0]+delay*y_0[1]]
    
    #Damping
    a_t = np.arange(0+delay, 1.0+delay, 0.2)
    y_after_delay=[sim_u_delay[-1],sim_u_dot_delay[-1]]    
    asol = integrate.odeint(ode_equ, y_after_delay, a_t)
    
    #print a_t_delay
    #print sim_u_delay
    #print sim_u_dot_delay
    #sys.exit()
    sim_a_t=np.concatenate((a_t_delay,a_t)).tolist()
    sim_u=np.concatenate((sim_u_delay,asol[:,0])).tolist()
    sim_u_dot=np.concatenate((sim_u_dot_delay,asol[:,1])).tolist()

    return sim_u[-1]






def control_turn(angle,u_global):
    global gyro,odo
    #Parameter
    eps_angle=1
    #Drive
    drive=l_drive.drive()
    duty=50

    
    if angle>=0:
        sgn=drive.turn_right(duty)
        odo.set_sign(sgn)     
    else:
        sgn=drive.turn_left(duty)
        odo.set_sign(sgn)

#     #Controlling    
#     a_l_last=0
#     a_r_last=0
#     
    #Define the variables arrays matrices
    #u_global=np.matrix(init_state).transpose()
    eps_odo=odo.rad_to_arc(1)
    t_global=[0]
    u_ref=np.matrix([0,0,0]).transpose()
    u_global_predicted=u_global
    enc_l_read_arr=[odo.rad_to_arc(odo.enc_l.value)]
    enc_r_read_arr=[odo.rad_to_arc(odo.enc_r.value)]
    enc_l_arr=[odo.rad_to_arc(odo.enc_l.value)]
    enc_r_arr=[odo.rad_to_arc(odo.enc_r.value)]
    yaw_vector=[math.radians(gyro.yaw.value)]
    #Fadeout
    fo_start=False
    fo_time=0.5
    print "######################Main loop###################################"
    print "#                                                                #"
    print "#                                                                #"
    print "#                                                                #"
    print "#                                                                #"
    print "##################################################################"
    i=1
    
    
    #Control circuit
    while True:
       ####################################
        #Position prediction
        ####################################
        #Time label of the measurement
        #print odo.t_label.value
        t_label=odo.t_label.value
        #Odometer
        enc_l_read_arr.append(odo.rad_to_arc(odo.enc_l.value))
        enc_r_read_arr.append(odo.rad_to_arc(odo.enc_r.value))
        #Read gyro
        yaw_vector.append(math.radians(gyro.yaw.value))
        d_yaw=yaw_vector[-1]-yaw_vector[-2]
#         print d_yaw
        #yaw_dot=gyro.yaw_dot.value
        t_global.append(t_label)
        #Get the displacement in the local coordinates: Only by Odo for check reasons
        [r_odo,d_alpha_odo]=get_arc(enc_l_read_arr[-1]-enc_l_read_arr[-2],enc_r_read_arr[-1]-enc_r_read_arr[-2])
        #print "Arc odo",[r_odo,d_alpha_odo]
        if d_alpha_odo!=0:
            du_local=get_local_disp(r_odo,d_alpha_odo,False)
        elif d_alpha_odo==0:
            du_local=get_local_disp(0,d_alpha_odo,enc_l_read_arr[-1]-enc_l_read_arr[-2])
            
        if True:
            #Prediction
            print "#############\nPredicted"
            if d_yaw!=0:            
                #Radius range
                [r_min,r_max]=get_radius_range(d_yaw,enc_l_read_arr[-1]-enc_l_arr[-1],enc_r_read_arr[-1]-enc_r_arr[-1],eps_odo)
                #Predict radius        
                r_predicted=(r_max+r_min)/2
                #Get predicted arc lengths
                [enc_l_predicted,enc_r_predicted]=get_arc_lengths(d_yaw,r_predicted)
                print "r_min=%.3f\tr_max=%.3f\tR Predicted:%.3f"%(r_min,r_max,r_predicted)
                #Get local displacements by this arcs
                du_local_predicted=get_local_disp(r_predicted,d_yaw,False)
                
            elif d_yaw==0:
                #Prediction
                #Radius range
                [l_min,l_max]=get_line_range(enc_l_read_arr[-1]-enc_l_arr[-1],enc_r_read_arr[-1]-enc_r_arr[-1],eps_odo)
                #Predict line length         
                enc_l_predicted=(l_max+l_min)/2
                enc_r_predicted=enc_l_predicted
                #Get local displacements by this arcs
                du_local_predicted=get_local_disp(0,0,enc_l_predicted)
                            #Predicted encoder arrays
                            
            enc_l_arr.append(enc_l_arr[-1]+enc_l_predicted)
            enc_r_arr.append(enc_r_arr[-1]+enc_r_predicted)
            
            #print "Predicted local at dyaw=%.3f"%d_yaw
            print du_local_predicted
        
        #################################
        #Reconstruct in the global system
        #################################
        
        u_global=update_global_disp(u_global,du_local)
        u_global_predicted=update_global_disp(u_global_predicted,du_local_predicted)

        
        u_ref=update_global_disp(u_ref,du_local_predicted)
        print u_ref
        yaw=u_ref[2,-1]
        #print (t_global[-1]-t_global[-2])    
        if (t_global[-1]-t_global[-2])!=0:
            yaw_dot=(u_ref[2,-1]-u_ref[2,-2])/(t_global[-1]-t_global[-2])    
        else:
            yaw_dot=0    
        rm=co_simulation_turn([yaw,yaw_dot])
                
        #print "Odo:%.4f\tYaw:%.4f\tYaw_dot:%.4f\tTarget distance:%.4f"%(rot_odo,yaw,yaw_dot,rm)
        #Sensor fusioning dummy
        #rot_act=0*rot_odo+1*rot_gyro
        #Break if x exceeds the value
        #print rot_act
        if math.radians(math.fabs(angle)-eps_angle)<=math.fabs(rm):
            print "Breaking"
            drive.break_move()            

            if not fo_start:
                print "##############################"
                fo_start=t_label
                
        if fo_start:
            if t_label-fo_start>=fo_time:
                #gyro.end()
                time.sleep(0.5)
                break
     
    
        #Clocking the controlling
        time.sleep(0.05)
    #print "Killing"
    #Kill subprocess
    #gyro.kill()
    #odo.kill()
    #One piece arc
    #du_global_op=get_local_displacement(enc_l[-1],enc_r[-1])
    #print du_global_op
    
    #print gyro.hist
    #############################################
    #Monitoring#
    #############################################
    print "Global position"
    print u_global_predicted[:,-1]
#     print "Global position one piece arc"
#     print du_global_op
    
    return [t_global,u_global_predicted]
    
    
    
    
    
#########################################################################
# Script
#########################################################################
#Start processes
#Gyrometer
gyro=l_gyro.gyro_asynch()
gyro.start()
#Odometry
odo=l_odo.odo_asynch() #Define object
odo.start(False) #Start measuring with the frequency

#Sequences
hist=control_straight(0.6,np.matrix([0,0,0]).transpose())
# hist=control_turn(90,hist[1])
# hist=control_straight(0.3,hist[1])
#hist=control_turn(90,np.matrix([1,1,math.pi/2.]).transpose())

odo.kill()
gyro.kill()









if False:
    t=hist[0]
    yaw_odo=hist[1][2,:].tolist()[0]
    yaw_gyro=hist[2]
#     print t
#     print yaw_odo
    plt.plot(t,yaw_odo)
    plt.plot(t,yaw_gyro)
    plt.show()



#Turn
if False:
    #print len(sim_u)
    #print len(sim_a_t)
    #plt.plot(sim_a_t,sim_u_dot)
    plt.plot(sim_a_t,sim_u)
    plt.show()
        
# if False:
#     arr_t=hist[0]
#     u_glob=hist[1]
#     
#     gyro_glob=hist[2]
#     #print u_glob
#     arr_alpha=u_glob[2,:].tolist()[0]
#     arr_yaw=gyro_glob[0,:].tolist()[0]
#     #print arr_alpha
#     l_misc.plots().write_gnu_file("odometry.dat",[arr_t,arr_alpha,arr_yaw],"%.6f")
#     
#     os.system("gnuplot -e \"plot  'odometry.dat' using 1:2 with lines, 'odometry.dat' using 1:3 with lines;pause -1\"")
#          
# if False:
#     gyrohist=hist[2]
#     arr_t=gyrohist[0]
#     arr_yaw=gyrohist[1]
#     arr_yaw_dot=gyrohist[2]
#     l_misc.plots().write_gnu_file("odometry.dat",[arr_t,arr_yaw,arr_yaw_dot],"%.6f")
#     os.system("gnuplot -e \"plot  'odometry.dat' using 1:2 with lines, 'odometry.dat' using 1:3 with lines ;pause -1\"")
#     
# 
#     
#     

if False:
    t_arr=hist[2][0]
    yaw_arr=hist[2][1]
    yaw_dot_arr=hist[2][2]
    t_odo=hist[0]
    yaw_odo=(hist[1][2,:]*180./math.pi).tolist()[0]
    #print yaw_odo
    fig, ax1 = plt.subplots()
    
    ax1.plot(t_arr, yaw_arr,"b-")
    #plt.plot([1,2,3,4], [0.1,0.4,0.9,0.16])
    
    ax2 = ax1.twinx()
    ax2.plot(t_arr, yaw_dot_arr,"r-")
    
    #ax3 = ax1.twinx()
    ax1.plot(t_odo,yaw_odo,"g-")
    plt.show()
    
    