#!/usr/bin/python
#############################################################
#############################################################
# Auto Drive
# Author: Can Kosar
#############################################################

#Standard libraries
import sys
import os
import time
import math
import numpy as np
import pygame
import sys

#Custom libraries
sys.path.append("../lib/")
import l_drive
import l_odo
import l_config
import l_monitoring
import l_misc

#Initializing
config=l_config.config
d=config.hardware.wheel_distance

#Get the displacement caused by arc motion
def get_local_displacement(d_a_l,d_a_r):
    if d_a_l!=d_a_r:
        r=(d*(d_a_l+d_a_r))/(2.*(d_a_l-d_a_r))
        
        d_alpha_hat=(d_a_l-d_a_r)/d
        dx_hat=r*math.sin(d_alpha_hat)
        dy_hat=r*(1-math.cos(d_alpha_hat))
    else:
        r=float("inf")
        dx_hat=d_a_l
        dy_hat=0
        d_alpha_hat=0
    
    if True:
        #print "################################   Turn ",i," t=",t_label,"######################################"
        #print "Actual enc_l= ",enc_l[-1]," Actual enc_r= ",enc_r[-1]
        print "Diff enc_l= ",d_a_l," Diff enc_r= ",d_a_r
        print "Actual radius= ",r
        print "Diff arc angle= ",d_alpha_hat,"(",math.degrees(d_alpha_hat),")", "Diff x local= ",dx_hat," Diff y local= ",dy_hat
        print "-------------Kinematics equations-------------"
    
    #Local coordinate system
    return np.matrix([dx_hat,dy_hat,d_alpha_hat]).transpose()

def control_straight(dist):
    #Odometry
    odo=l_odo.odo_asynch() #Define object
    odo.start(False) #Start measuring with the frequency
    #Drive       
    drive=l_drive.drive()
    duty=70
    balance_calibration=0
    sgn=drive.move_fw(duty,balance_calibration)
    odo.set_sign(sgn)
    
    
    
     
    #Controlling
    
    a_l_last=0
    a_r_last=0
    
    #Define the variables arrays matrices
    u_global=np.matrix([0,0,0]).transpose()
    enc_l=[0]
    enc_r=[0]
    t_vector=[0]
    
    print "######################Main loop###################################"
    print "#                                                                #"
    print "#                                                                #"
    print "#                                                                #"
    print "#                                                                #"
    print "##################################################################"
    i=1
    
    
    #Control circuit
    while True:
        ####################################
        #Position prediction
        ####################################
        enc_l.append(odo.rad_to_arc(odo.enc_l.value))
        enc_r.append(odo.rad_to_arc(odo.enc_r.value))
        #Time label of the measurement
        t_label=odo.t_label.value
        t_vector.append(t_label)
        #Monitoring
        print "################################   Turn ",i," t=",t_label,"######################################"
        #Get the displacement in the local coordinates
        du_local=get_local_displacement(enc_l[-1]-enc_l[-2],enc_r[-1]-enc_r[-2])
        #Rotation of the local system regarding the global system 
        d_alpha_gl=u_global[2,-1]
        #Transformation-matrix (Global to local)
        T= np.matrix([[math.cos(d_alpha_gl), -math.sin(d_alpha_gl),0],
                      [math.sin(d_alpha_gl), math.cos(d_alpha_gl),0],
                      [0,0,1]])
        #Displacement in the global system
        du_global=T*du_local
        #Absolute position
        u_global=np.concatenate((u_global,u_global[:,-1]+du_global),axis=1)
    
        
        #Calculating the controlling parameters
        
        #Error vector (In this case it's y-displacement: e(t)=u_G,y(t)
        e=u_global[:,-1]
        t=t_vector[-1]
        if t-t_vector[-2]>0:
            e_dot=(e-u_global[1,-2])/(t-t_vector[-2])
        else:
            e_dot=np.matrix([0,0,0]).transpose()
            print "Warning: Time increment unsuccessfull"
        
        #State space
        ss=np.concatenate((e,e_dot),axis=0).transpose()
        #Parameter matrix    
        A=np.matrix([0,-100,-150,0,-10,-0.5]).transpose()
        #Set matrix
        print l_misc.plots().print_matrix(ss)
        print l_misc.plots().print_matrix(A)
        sm=ss*A
        
        
        #PD-Controller
        #print l_misc.plots().print_matrix(e)
        
        new_balance=sm[0]
        
        sgn=drive.move_fw(duty,balance_calibration+new_balance)
        
        #############################################
        #Monitoring#
        #############################################
        if True:
            print "-------Controlling---------"
            print "e"
            print l_misc.plots().print_matrix(e)
            print "e dot"
            print l_misc.plots().print_matrix(e_dot)
            print "New balance=%.3f"%new_balance
            print "---------------------------"
            print "du_local"
            print l_misc.plots().print_matrix(du_local)
            print "T"
            print l_misc.plots().print_matrix(T)
            print "du_global"
            print l_misc.plots().print_matrix(du_global)
            print "u_global"
            print l_misc.plots().print_matrix(u_global)
            print "\n\n\n"
        #############################################
        #Exit criteria    
        i+=1
        #Actual x position    
        x_act=u_global[0,-1]
        #Break if x exceeds the value
        if x_act>=dist:
            drive.break_move()
            odo.kill()
            break
        
    
        #Clocking the controlling
        time.sleep(0.01)
    
    #One piece arc
    du_global_op=get_local_displacement(enc_l[-1],enc_r[-1])
    
    
    #############################################
    #Monitoring#
    #############################################
    print "Global position"
    print u_global[:,-1]
    print "Global position one piece arc"
    print du_global_op




def control_turn(angle):
    
    #Odometry
    odo=l_odo.odo_asynch() #Define object
    odo.start(False) #Start measuring with the frequency
    #Drive
    drive=l_drive.drive()
    duty=50
    sgn=drive.turn_right(duty)
    odo.set_sign(sgn)
    
    
    
     
    #Controlling
    
    a_l_last=0
    a_r_last=0
    
    #Define the variables arrays matrices
    u_global=np.matrix([0,0,0]).transpose()
    enc_l=[0]
    enc_r=[0]
    t_vector=[0]
    
    print "######################Main loop###################################"
    print "#                                                                #"
    print "#                                                                #"
    print "#                                                                #"
    print "#                                                                #"
    print "##################################################################"
    i=1
    
    
    #Control circuit
    while True:
        ####################################
        #Position prediction
        ####################################
        enc_l.append(odo.rad_to_arc(odo.enc_l.value))
        enc_r.append(odo.rad_to_arc(odo.enc_r.value))
        #Time label of the measurement
        t_label=odo.t_label.value
        t_vector.append(t_label)
        #Get the displacement in the local coordinates
        du_local=get_local_displacement(enc_l[-1]-enc_l[-2],enc_r[-1]-enc_r[-2])
        #Rotation of the local system regarding the global system 
        d_alpha_gl=u_global[2,-1]
        #Transformation-matrix (Global to local)
        T= np.matrix([[math.cos(d_alpha_gl), -math.sin(d_alpha_gl),0],
                      [math.sin(d_alpha_gl), math.cos(d_alpha_gl),0],
                      [0,0,1]])
        #Displacement in the global system
        du_global=T*du_local
        #Absolute position
        u_global=np.concatenate((u_global,u_global[:,-1]+du_global),axis=1)
    
        #Exit criteria    
        i+=1
        #Actual x position    
        rot_act=math.degrees(u_global[2,-1])
        #Break if x exceeds the value
        if rot_act>=angle:
            drive.break_move()
            odo.kill()
            break
        
    
        #Clocking the controlling
        time.sleep(0.01)
    
    #One piece arc
    du_global_op=get_local_displacement(enc_l[-1],enc_r[-1])
    
    
    #############################################
    #Monitoring#
    #############################################
    print "Global position"
    print u_global[:,-1]
    print "Global position one piece arc"
    print du_global_op
    
    
    
control_straight(0.70)
time.sleep(0.5)
control_turn(90)
time.sleep(0.5)
control_straight(0.70)
time.sleep(0.5)
control_turn(90)
time.sleep(0.5)
control_straight(0.70)
time.sleep(0.5)
control_turn(90)
time.sleep(0.5)
control_straight(0.70)
time.sleep(0.5)
control_turn(90)